# TextParse

TextParse uses Julia's generated functions to generate efficient specialized parsers for text files. TextParse minimizes allocations and hence avoids involving the GC.

```@contents
```

## Installation

```julia
Pkg.add("TextParse")
```

## Reading CSV

The most useful API is probably `csvread` - read a CSV file:

```@docs
csvread
```

Some notable features of the CSV parser are:

- CSV parsing kernel generated by TextParse is type-inferable.
- Uses PooledArrays for strings by default, promoting to an `Array` only if number of unique elements goes over 5% (after 400 rows have been read).
- Avoids allocating the string in PooledArray unless the string is not in the pool
- Doesn't assume all columns are nullable by default, switches column to `DataValueArray` if an NA value is found
- Flexible about predicted column types, can convert the column mid-way if the type changes and switch to a new fast generated method
- Fast date time parsing even on Julia 0.5

## Extensible parsing framework

TextParse operates by defining small parsers which are specialized to parse one kind of text very efficiently. Each such parser is described by a subtype of `AbstractToken{T}`. An `AbstractToken{T}` type should implement a `tryparsenext` method:

```@docs
TextParse.tryparsenext
```

### Available AbstractToken types

```@docs
TextParse.Numeric
```

```@docs
TextParse.StringToken
```

```@docs
TextParse.DateTimeToken
```

```@docs
TextParse.NAToken
```
```@docs
TextParse.Quoted
```

```@docs
CustomParser
```

### LocalOpts

```@docs
TextParse.LocalOpts
```
